"""Cyclopts central Group/Parameter resolver.

All fallbacks and configuration hierarchy resolution should occur here.
All downstream functions should consume data "as is" without fallbacks.
"""

import inspect
from functools import cached_property
from typing import Any, Callable, Dict, List, Optional, Tuple, cast, get_origin

import cyclopts.utils
from cyclopts.exceptions import DocstringError
from cyclopts.group import Group
from cyclopts.parameter import Parameter, get_hint_parameter
from cyclopts.utils import ParameterDict


def _list_index(lst: List, key: Callable) -> int:
    """Returns index of first occurrence in list."""
    for i, element in enumerate(lst):
        if key(element):
            return i
    raise ValueError


def _has_unparsed_parameters(func_signature: inspect.Signature, *args) -> bool:
    for iparam in func_signature.parameters.values():
        cparam: Parameter
        _, cparam = get_hint_parameter(iparam, *args)
        if not cparam.parse:
            return True
    return False


def _resolve_groups(
    func_signature: inspect.Signature,
    app_parameter: Optional[Parameter],
    group_arguments: Group,
    group_parameters: Group,
):
    """Resolves groups and mapping iparams to groups.

    cparams will have to be externally re-resolved to include group.default_parameter
    """
    resolved_groups = []
    iparam_to_groups = ParameterDict()

    for iparam in func_signature.parameters.values():
        _, cparam = get_hint_parameter(iparam, app_parameter)

        if not cparam.parse:
            continue

        if cparam.group:
            groups = cparam.group
        elif iparam.kind is iparam.POSITIONAL_ONLY:
            groups = (group_arguments,)
        else:
            groups = (group_parameters,)

        iparam_to_groups.setdefault(iparam, [])

        for group in groups:  # pyright: ignore
            if isinstance(group, str):
                try:
                    index = _list_index(resolved_groups, lambda x: x.name == group)  # noqa: B023
                except ValueError:
                    group = Group(group)
                    resolved_groups.append(group)
                else:
                    group = resolved_groups[index]
                iparam_to_groups[iparam].append(group)
            elif isinstance(group, Group):
                # Ensure a different, but same-named group doesn't already exist
                if any(group is not x and x.name == group.name for x in resolved_groups):
                    raise ValueError("Cannot register 2 distinct Group objects with same name.")

                if group.default_parameter is not None and group.default_parameter.group:
                    # This shouldn't be possible due to ``Group`` internal checks.
                    raise ValueError("Group.default_parameter cannot have a specified group.")  # pragma: no cover

                try:
                    index = resolved_groups.index(group)
                except ValueError:
                    resolved_groups.append(group)
                else:
                    group = resolved_groups[index]
                iparam_to_groups[iparam].append(group)
            else:
                raise TypeError

    return resolved_groups, iparam_to_groups


def _resolve_docstring(f: Callable, signature: inspect.Signature) -> ParameterDict:
    from docstring_parser import parse as docstring_parse

    iparam_to_docstring_cparam = ParameterDict()
    if f.__doc__ is None:
        return iparam_to_docstring_cparam
    f_docstring = docstring_parse(f.__doc__)

    for dparam in f_docstring.params:
        try:
            iparam = signature.parameters[dparam.arg_name]
        except KeyError:
            # Even though we could pass/continue, we're raising
            # an exception because the developer really aught to know.
            raise DocstringError(
                f"Docstring parameter {dparam.arg_name} has no equivalent in function signature."
            ) from None
        else:
            iparam_to_docstring_cparam[iparam] = Parameter(help=dparam.description)

    return iparam_to_docstring_cparam


class ResolvedCommand:
    command: Callable
    groups: List[Group]
    groups_iparams: List[Tuple[Group, List[inspect.Parameter]]]
    iparam_to_groups: ParameterDict
    iparam_to_cparam: ParameterDict

    pyname_to_iparam: Dict[str, inspect.Parameter]
    # Plain python identifier string to inspect.Parameter

    def __init__(
        self,
        f,
        app_parameter: Optional[Parameter] = None,
        group_arguments: Optional[Group] = None,
        group_parameters: Optional[Group] = None,
        parse_docstring: bool = True,
    ):
        """
        ``app_parameter`` implicitly has the command-group parameter already resolved.

        Parameters
        ----------
        f: Callable
            Function to resolve annotated :class:`Parameters`.
        app_parameter:
            Default :class:`Parameter` to inherit configuration from.
        group_arguments: Optional[Group]
            Default :class:`Group` for positional-only arguments.
        group_parameters: Optional[Group]
            Default :class:`Group` for non-positional-only arguments.
        parse_docstring: bool
            Parse the docstring to populate Parameter ``help``, if not explicitly set.
            Disable for improved performance if ``help`` won't be used in the resulting :class:`Parameter`.
        """
        if group_arguments is None:
            group_arguments = Group.create_default_arguments()
        if group_parameters is None:
            group_parameters = Group.create_default_parameters()

        self.command = f
        signature = cyclopts.utils.signature(f)
        self.pyname_to_iparam = cast(Dict[str, inspect.Parameter], signature.parameters)

        # Get:
        # 1. Fully resolved and created Groups.
        # 2. A mapping of inspect.Parameter to those Group objects.
        self.groups, self.iparam_to_groups = _resolve_groups(
            signature, app_parameter, group_arguments, group_parameters
        )

        # Fully Resolve each Cyclopts Parameter
        self.iparam_to_cparam = ParameterDict()
        iparam_to_docstring_cparam = _resolve_docstring(f, signature) if parse_docstring else ParameterDict()
        empty_help_string_parameter = Parameter(help="")
        for iparam, groups in self.iparam_to_groups.items():
            cparam = get_hint_parameter(
                iparam,
                empty_help_string_parameter,
                app_parameter,
                *(x.default_parameter for x in groups),
                iparam_to_docstring_cparam.get(iparam),
                Parameter(required=iparam.default is iparam.empty),
            )[1]

            # Resolve ``name`` now that ``name_transform`` has been resolved.
            if iparam.kind in (iparam.POSITIONAL_ONLY, iparam.VAR_POSITIONAL):
                # Name is only used for help-string
                names = [iparam.name.upper()]
            else:
                # cparam.name_transform cannot be None due to:
                #     attrs.converters.default_if_none(default_name_transform)
                assert cparam.name_transform is not None
                names = ["--" + cparam.name_transform(iparam.name)]

            cparam = Parameter.combine(Parameter(name=names), cparam)
            self.iparam_to_cparam[iparam] = cparam

        self.bind = signature.bind_partial

        # Create a convenient group-to-iparam structure
        self.groups_iparams = [
            (
                group,
                [iparam for iparam, groups in self.iparam_to_groups.items() if group in groups],
            )
            for group in self.groups
        ]

    @property
    def iparams(self):
        return self.iparam_to_cparam.keys()

    @cached_property
    def cli2parameter(self) -> Dict[str, Tuple[inspect.Parameter, Any]]:
        """Creates a dictionary mapping CLI keywords to python keywords.

        Typically the mapping is something like::

            {"--foo": (<Parameter "foo">, None)}

        Each value is a tuple containing:

        1. The corresponding ``inspect.Parameter``.
        2. A predefined value. If this value is :obj:`None`, the value should be
           inferred from subsequent tokens.
        """
        # The tuple's second element is an implicit value for flags.
        mapping: Dict[str, Tuple[inspect.Parameter, Any]] = {}

        for iparam, cparam in self.iparam_to_cparam.items():
            if iparam.kind is iparam.VAR_KEYWORD:
                # Don't directly expose the kwarg variable name
                continue
            hint = get_hint_parameter(iparam)[0]
            for name in cparam.name:
                mapping[name] = (iparam, True if hint is bool else None)
            for name in cparam.get_negatives(hint, *cparam.name):
                # Creates empty versions of iterables.
                mapping[name] = (iparam, (get_origin(hint) or hint)())

        return mapping

    @cached_property
    def parameter2cli(self) -> ParameterDict:
        """Creates a dictinoary mapping :class:`inspect.Parameter` to a list of CLI names."""
        c2p = self.cli2parameter
        p2c = ParameterDict()

        for cli, tup in c2p.items():
            iparam = tup[0]
            p2c.setdefault(iparam, [])
            p2c[iparam].append(cli)

        for iparam, cparam in self.iparam_to_cparam.items():
            # POSITIONAL_OR_KEYWORD and KEYWORD_ONLY already handled in cli2parameter
            if iparam.kind in (iparam.POSITIONAL_ONLY, iparam.VAR_KEYWORD, iparam.VAR_POSITIONAL):
                p2c[iparam] = list(cparam.name)

        return p2c



import functools
import inspect
import sys
from collections.abc import MutableMapping
from typing import (
    Any,
    Dict,
    Iterable,
    Iterator,
    List,
    Literal,
    Optional,
    Tuple,
    Type,
    Union,
)

_union_types = set()
_union_types.add(Union)
if sys.version_info >= (3, 10):
    from types import UnionType

    _union_types.add(UnionType)

# fmt: off
if sys.version_info >= (3, 10):
    def signature(f: Any) -> inspect.Signature:
        return inspect.signature(f, eval_str=True)
else:
    def signature(f: Any) -> inspect.Signature:
        return inspect.signature(f)
# fmt: on


class SentinelMeta(type):
    def __repr__(cls) -> str:
        return f"<{cls.__name__}>"

    def __bool__(cls) -> Literal[False]:
        return False


class Sentinel(metaclass=SentinelMeta):
    pass


def record_init(target: str):
    """Class decorator that records init argument names as a tuple to ``target``."""

    def decorator(cls):
        original_init = cls.__init__
        function_signature = signature(original_init)

        @functools.wraps(original_init)
        def new_init(self, *args, **kwargs):
            bound = function_signature.bind(self, *args, **kwargs)
            original_init(self, *args, **kwargs)
            # Circumvent frozen protection.
            object.__setattr__(self, target, tuple(k for k, v in bound.arguments.items() if v is not self))

        cls.__init__ = new_init
        return cls

    return decorator


def is_iterable(obj) -> bool:
    return isinstance(obj, Iterable) and not isinstance(obj, str)


def is_union(type_: Optional[Type]) -> bool:
    return type_ in _union_types


class ParameterDict(MutableMapping):
    """A dictionary implementation that can handle mutable ``inspect.Parameter`` as keys."""

    def __init__(self, store: Optional[Dict[inspect.Parameter, Any]] = None):
        self.store = {}
        self.reverse_mapping = {}
        if store is not None:
            for k, v in store.items():
                self[k] = v

    def _param_key(self, param: inspect.Parameter) -> tuple:
        if not isinstance(param, inspect.Parameter):
            raise TypeError(f"Key must be an inspect.Parameter; got {type(param)}.")
        return (param.name, param.kind, param.annotation)

    def __getitem__(self, key: inspect.Parameter) -> Any:
        return self.store[self._param_key(key)]

    def __setitem__(self, key: inspect.Parameter, value: Any) -> None:
        processed_key = self._param_key(key)
        self.store[processed_key] = value
        self.reverse_mapping[processed_key] = key

    def __delitem__(self, key: inspect.Parameter) -> None:
        processed_key = self._param_key(key)
        del self.store[processed_key]
        del self.reverse_mapping[processed_key]

    def __iter__(self) -> Iterator[inspect.Parameter]:
        return iter(self.reverse_mapping.values())

    def __len__(self) -> int:
        return len(self.store)

    def __repr__(self) -> str:
        inner = []
        for key, value in self.store.items():
            inner.append(f"Parameter(name={key[0]!r}, kind={key[1]}, annotation={key[2]}): {value}")
        return "{" + ", ".join(inner) + "}"

    def __contains__(self, key: object) -> bool:
        if not isinstance(key, inspect.Parameter):
            raise TypeError(f"Key must be an inspect.Parameter; got {type(key)}.")
        return self._param_key(key) in self.store

    def setdefault(self, key: inspect.Parameter, default: Any = None) -> Any:
        processed_key = self._param_key(key)
        if processed_key not in self.store:
            self.reverse_mapping[processed_key] = key
        return self.store.setdefault(processed_key, default)

    def get(self, key: inspect.Parameter, default: Any = None):
        try:
            return self[key]
        except KeyError:
            return default

    def clear(self):
        self.store.clear()
        self.reverse_mapping.clear()


def resolve_callables(t, *args, **kwargs):
    """Recursively resolves callable elements in a tuple."""
    if isinstance(t, type(Sentinel)):
        return t

    if callable(t):
        return t(*args, **kwargs)

    resolved = []
    for element in t:
        if isinstance(element, type(Sentinel)):
            resolved.append(element)
        elif callable(element):
            resolved.append(element(*args, **kwargs))
        elif is_iterable(element):
            resolved.append(resolve_callables(element, *args, **kwargs))
        else:
            resolved.append(element)
    return tuple(resolved)


def to_tuple_converter(value: Union[None, Any, Iterable[Any]]) -> Tuple[Any, ...]:
    """Convert a single element or an iterable of elements into a tuple.

    Intended to be used in an ``attrs.Field``. If :obj:`None` is provided, returns an empty tuple.
    If a single element is provided, returns a tuple containing just that element.
    If an iterable is provided, converts it into a tuple.

    Parameters
    ----------
    value: Optional[Union[Any, Iterable[Any]]]
        An element, an iterable of elements, or None.

    Returns
    -------
    Tuple[Any, ...]: A tuple containing the elements.
    """
    if value is None:
        return ()
    elif is_iterable(value):
        return tuple(value)
    else:
        return (value,)


def to_list_converter(value: Union[None, Any, Iterable[Any]]) -> List[Any]:
    return list(to_tuple_converter(value))


def optional_to_tuple_converter(value: Union[None, Any, Iterable[Any]]) -> Optional[Tuple[Any, ...]]:
    """Convert a string or Iterable or None into an Iterable or None.

    Intended to be used in an ``attrs.Field``.
    """
    if value is None:
        return None

    if not value:
        return ()

    return to_tuple_converter(value)


def default_name_transform(s: str):
    """Converts a python identifier into a CLI token.

    Performs the following operations (in order):

    1. Convert the string to all lowercase.
    2. Replace ``_`` with ``-``.
    3. Strip any leading/trailing ``-`` (also stripping ``_``, due to point 2).

    Intended to be used with :attr:`App.name_transform` and :attr:`Parameter.name_transform`.

    Parameters
    ----------
    s: str
        Input python identifier string.

    Returns
    -------
    str
        Transformed name.
    """
    return s.lower().replace("_", "-").strip("-")



import errno
import inspect
import os
from abc import ABC, abstractmethod
from contextlib import suppress
from itertools import chain
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Optional, Set, Tuple, Union

from attrs import define, field

from cyclopts.exceptions import UnknownOptionError
from cyclopts.utils import to_tuple_converter

if TYPE_CHECKING:
    from cyclopts.core import App


@define
class Unset:
    """Placeholder object for a parameter that does not yet have any associated string tokens.

    Used with :attr:`App.config <cyclopts.App.config>`.

    Parameters
    ----------
    iparam: inspect.Parameter
        The corresponding :class:`inspect.Parameter` for the unset variable.
    related: set[str]
        Other CLI names that map to the same :class:`inspect.Parameter`.
        These may be aliases, or things like negative flags.
    """

    iparam: inspect.Parameter
    related: Set[str] = field(factory=set)

    def related_set(self, mapping: Dict[str, Union["Unset", List[str]]]) -> Set[str]:
        """Other CLI keys that map to the same :class:`inspect.Parameter` that have parsed token(s).

        Parameters
        ----------
        mapping: dict
            All associated cli_name to tokens.

        Returns
        -------
        set[str]
            CLI keys that map to the same :class:`inspect.Parameter` that have parsed token(s).
        """
        return {x for x in self.related.intersection(mapping) if not isinstance(mapping[x], Unset)}


@define
class ConfigFromFile(ABC):
    path: Union[str, Path] = field(converter=Path)
    root_keys: Iterable[str] = field(default=(), converter=to_tuple_converter)
    must_exist: bool = field(default=False, kw_only=True)
    search_parents: bool = field(default=False, kw_only=True)
    allow_unknown: bool = field(default=False, kw_only=True)

    _config: Optional[Dict[str, Any]] = field(default=None, init=False, repr=False)

    @abstractmethod
    def _load_config(self, path: Path) -> Dict[str, Any]:
        """Load the config dictionary from path.

        Parameters
        ----------
        path: Path
            Path to the file. Guaranteed to exist.

        Returns
        -------
        dict
            Loaded configuration.
        """
        raise NotImplementedError

    @property
    def config(self) -> Dict[str, Any]:
        if self._config is not None:
            return self._config

        assert isinstance(self.path, Path)
        for parent in self.path.parents:
            candidate = parent / self.path.name
            if candidate.exists():
                self._config = self._load_config(candidate)
                return self._config
            elif self.search_parents:
                # Continue iterating over parents.
                continue
            elif self.must_exist:
                raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), str(self.path))

        # No matching file was found.
        if self.must_exist:
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), str(self.path))

        self._config = {}
        return self._config

    def __call__(self, apps: List["App"], commands: Tuple[str, ...], mapping: Dict[str, Union[Unset, List[str]]]):
        config = self.config
        try:
            for key in chain(self.root_keys, commands):
                config = config[key]
        except KeyError:
            return

        if not self.allow_unknown and (remaining_keys := set(config) - set(apps[-1]) - set(mapping)):
            raise UnknownOptionError(token=sorted(remaining_keys)[0])

        for key, value in mapping.items():
            if not isinstance(value, Unset) or value.related_set(mapping):
                continue

            with suppress(KeyError):
                new_value = config[key]
                if not isinstance(new_value, list):
                    new_value = [new_value]
                mapping[key] = [str(x) for x in new_value]



from typing import Optional


class LimitedChoice:
    def __init__(self, min: int = 0, max: Optional[int] = None):
        """Group validator that limits the number of selections per group.

        Commonly used for enforcing mutually-exclusive parameters (default behavior).

        Parameters
        ----------
        min: int
            The minimum (inclusive) number of CLI parameters allowed.
        max: Optional[int]
            The maximum (inclusive) number of CLI parameters allowed.
            Defaults to ``1`` if ``min==0``, ``min`` otherwise.
        """
        self.min = min
        self.max = (self.min or 1) if max is None else max
        if self.max < self.min:
            raise ValueError("max must be >=min.")

    def __call__(self, **kwargs):
        if not (self.min <= len(kwargs) <= self.max):
            offenders = "{" + ", ".join(f"--{k}" for k in kwargs.keys()) + "}"
            if self.min == 0 and self.max == 1:
                raise ValueError(f"Mutually exclusive arguments: {offenders}")
            else:
                raise ValueError(
                    f"Received {len(kwargs)} arguments: {offenders}. Only [{self.min}, {self.max}] choices may be specified."
                )



import itertools
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Iterable,
    List,
    Optional,
    Tuple,
    Union,
    cast,
)

from attrs import define, field

from cyclopts.utils import Sentinel, is_iterable, resolve_callables, to_tuple_converter

if TYPE_CHECKING:
    from cyclopts.parameter import Parameter


def _group_default_parameter_must_be_none(instance, attribute, value: Optional["Parameter"]):
    if value is None:
        return

    if value.group:
        raise ValueError("Group default_parameter cannot have a group.")


# Used for Group.sorted
_sort_key_counter = itertools.count()


class NO_USER_SORT_KEY(Sentinel):  # noqa: N801
    pass


@define
class Group:
    name: str = ""

    help: str = ""

    # All below parameters are keyword-only
    _show: Optional[bool] = field(default=None, alias="show", kw_only=True)

    _sort_key: Any = field(
        default=None,
        alias="sort_key",
        converter=lambda x: NO_USER_SORT_KEY if x is None else x,
    )

    converter: Optional[Callable] = field(default=None, kw_only=True)

    # This can ONLY ever be a Tuple[Callable, ...]
    validator: Union[None, Callable, Iterable[Callable]] = field(
        default=None,
        converter=lambda x: cast(Tuple[Callable, ...], to_tuple_converter(x)),
        kw_only=True,
    )

    default_parameter: Optional["Parameter"] = field(
        default=None,
        validator=_group_default_parameter_must_be_none,
        kw_only=True,
    )

    def __str__(self):
        return self.name

    @property
    def show(self):
        return bool(self.name) if self._show is None else self._show

    @show.setter
    def show(self, value):
        self._show = value

    @property
    def sort_key(self):
        return None if self._sort_key is NO_USER_SORT_KEY else self._sort_key

    @sort_key.setter
    def sort_key(self, value):
        self._sort_key = value

    @classmethod
    def create_default_arguments(cls):
        return cls("Arguments")

    @classmethod
    def create_default_parameters(cls):
        return cls("Parameters")

    @classmethod
    def create_default_commands(cls):
        return cls("Commands")

    @classmethod
    def create_ordered(cls, *args, sort_key=None, **kwargs):
        """Create a group with a globally incremented :attr:`~Group.sort_key`.

        Used to create a group that will be displayed **after** a previously declared :meth:`Group.create_ordered` group on the help-page.

        If a :attr:`~Group.sort_key` is provided, it is **prepended** to the globally incremented counter value (i.e. has priority during sorting).
        """
        count = next(_sort_key_counter)
        if sort_key is None:
            sort_key = (NO_USER_SORT_KEY, count)
        elif is_iterable(sort_key):
            sort_key = (tuple(sort_key), count)
        else:
            sort_key = (sort_key, count)
        return cls(*args, sort_key=sort_key, **kwargs)


@define
class GroupConverter:
    default_group: Group

    def __call__(self, input_value: Union[None, str, Group]) -> Group:
        if input_value is None:
            return self.default_group
        elif isinstance(input_value, str):
            return Group(input_value)
        elif isinstance(input_value, Group):
            return input_value
        else:
            raise TypeError


def sort_groups(groups: List[Group], attributes: List[Any]) -> Tuple[List[Group], List[Any]]:
    """Sort groups for the help-page."""
    assert len(groups) == len(attributes)
    if not groups:
        return groups, attributes

    # Resolve callable ``sort_key``
    sort_key__group_attributes = []
    for group, attribute in zip(groups, attributes):
        value = (group, attribute)
        if callable(group._sort_key) or is_iterable(group._sort_key):
            sort_key__group_attributes.append((resolve_callables(group._sort_key, group), value))
        else:
            sort_key__group_attributes.append((group._sort_key, value))

    sort_key_panels: List[Tuple[Tuple, Tuple[Group, Any]]] = []
    ordered_no_user_sort_key_panels: List[Tuple[Tuple, Tuple[Group, Any]]] = []
    no_user_sort_key_panels: List[Tuple[Tuple, Tuple[Group, Any]]] = []

    for sort_key, (group, attribute) in sort_key__group_attributes:
        value = (group, attribute)
        if sort_key in (NO_USER_SORT_KEY, None):
            no_user_sort_key_panels.append(((group.name,), value))
        elif is_iterable(sort_key) and sort_key[0] in (NO_USER_SORT_KEY, None):
            ordered_no_user_sort_key_panels.append((sort_key[1:] + (group.name,), value))
        else:
            sort_key_panels.append(((sort_key, group.name), value))

    sort_key_panels.sort()
    ordered_no_user_sort_key_panels.sort()
    no_user_sort_key_panels.sort()

    combined = sort_key_panels + ordered_no_user_sort_key_panels + no_user_sort_key_panels

    out_groups, out_attributes = zip(*[x[1] for x in combined])

    return list(out_groups), list(out_attributes)



import inspect
import sys
from enum import Enum
from functools import lru_cache, partial
from inspect import isclass
from typing import (
    TYPE_CHECKING,
    Callable,
    Iterable,
    List,
    Literal,
    Optional,
    Tuple,
    Type,
    Union,
    get_args,
    get_origin,
)

from attrs import define, field, frozen

import cyclopts.utils
from cyclopts.group import Group
from cyclopts.parameter import Parameter, get_hint_parameter

if TYPE_CHECKING:
    from rich.console import RenderableType

    from cyclopts.core import App

if sys.version_info >= (3, 12):
    from typing import TypeAliasType
else:
    TypeAliasType = None


@lru_cache(maxsize=16)
def docstring_parse(doc: str):
    """Addon to :func:`docstring_parser.parse` that double checks the `short_description`."""
    import docstring_parser

    res = docstring_parser.parse(doc)
    cleaned_doc = inspect.cleandoc(doc)
    short = cleaned_doc.split("\n\n")[0]
    if res.short_description != short:
        if res.long_description is None:
            res.long_description = res.short_description
        elif res.short_description is not None:
            res.long_description = res.short_description + "\n" + res.long_description
        res.short_description = None
    return res


@frozen
class HelpEntry:
    name: str
    short: str
    description: "RenderableType"
    required: bool = False


def _text_factory():
    from rich.text import Text

    return Text()


@define
class HelpPanel:
    format: Literal["command", "parameter"]
    title: str
    description: "RenderableType" = field(factory=_text_factory)
    entries: List[HelpEntry] = field(factory=list)

    def remove_duplicates(self):
        seen, out = set(), []
        for item in self.entries:
            hashable = (item.name, item.short)
            if hashable not in seen:
                seen.add(hashable)
                out.append(item)
        self.entries = out

    def sort(self):
        self.entries.sort(key=lambda x: (x.name.startswith("-"), x.name))

    def __rich__(self):
        if not self.entries:
            return _silent
        from rich.box import ROUNDED
        from rich.console import Group as RichGroup
        from rich.console import NewLine
        from rich.panel import Panel
        from rich.table import Table
        from rich.text import Text

        table = Table.grid(padding=(0, 1))
        panel_description = self.description

        if isinstance(panel_description, Text):
            panel_description.end = ""

            if panel_description.plain:
                panel_description = RichGroup(panel_description, NewLine(2))
        else:
            # Should be either a RST or Markdown object
            if panel_description.markup:  # pyright: ignore[reportAttributeAccessIssue]
                panel_description = RichGroup(panel_description, NewLine(1))

        panel = Panel(
            RichGroup(panel_description, table),
            box=ROUNDED,
            expand=True,
            title_align="left",
            title=self.title,
        )

        if self.format == "command":
            table.add_column(justify="left", style="cyan")
            table.add_column(justify="left")

            for entry in self.entries:
                name = entry.name
                if entry.short:
                    name += "," + entry.short
                table.add_row(name + " ", entry.description)
        elif self.format == "parameter":
            has_short = any(entry.short for entry in self.entries)
            has_required = any(entry.required for entry in self.entries)

            if has_required:
                table.add_column(justify="left", width=1, style="red bold")  # For asterisk
            table.add_column(justify="left", no_wrap=True, style="cyan")  # For option names
            if has_short:
                table.add_column(justify="left", no_wrap=True, style="green")  # For short options
            table.add_column(justify="left")  # For main help text.

            for entry in self.entries:
                row = []
                if has_required:
                    if entry.required:
                        row.append("*")
                    else:
                        row.append("")
                row.append(entry.name + " ")
                if has_short:
                    row.append(entry.short + " ")
                row.append(entry.description)
                table.add_row(*row)
        else:
            raise NotImplementedError

        return panel


class SilentRich:
    """Dummy object that causes nothing to be printed."""

    def __rich_console__(self, console, options):
        # This generator yields nothing, so ``rich`` will print nothing for this object.
        if False:
            yield


_silent = SilentRich()


def _is_short(s):
    return not s.startswith("--") and s.startswith("-")


def format_usage(
    app,
    command_chain: Iterable[str],
):
    from rich.text import Text

    usage = []
    usage.append("Usage:")
    usage.append(app.name[0])
    usage.extend(command_chain)

    for command in command_chain:
        app = app[command]

    if app._commands:
        usage.append("COMMAND")

    if app.default_command:
        to_show = set()
        for parameter in cyclopts.utils.signature(app.default_command).parameters.values():
            if parameter.kind in (parameter.POSITIONAL_ONLY, parameter.VAR_POSITIONAL, parameter.POSITIONAL_OR_KEYWORD):
                to_show.add("[ARGS]")
            if parameter.kind in (parameter.KEYWORD_ONLY, parameter.VAR_KEYWORD, parameter.POSITIONAL_OR_KEYWORD):
                to_show.add("[OPTIONS]")
        usage.extend(sorted(to_show))

    return Text(" ".join(usage) + "\n", style="bold")


def format_doc(root_app, app: "App", format: str = "restructuredtext"):
    from rich.console import Group as RichGroup
    from rich.console import NewLine
    from rich.text import Text

    raw_doc_string = app.help

    if not raw_doc_string:
        return _silent

    parsed = docstring_parse(raw_doc_string)

    components: List[Union[str, Tuple[str, str]]] = []
    if parsed.short_description:
        components.append(parsed.short_description + "\n")

    if parsed.long_description:
        if parsed.short_description:
            components.append("\n")
        components.append((parsed.long_description + "\n", "info"))

    return RichGroup(format_str(*components, format=format), NewLine())


def format_str(*components: Union[str, Tuple[str, str]], format: str = "restructuredtext") -> "RenderableType":
    format = format.lower()

    if format == "plaintext":
        from rich.text import Text

        aggregate = []
        for component in components:
            if isinstance(component, str):
                aggregate.append(component)
            else:
                aggregate.append(component[0])
        return Text.assemble("".join(aggregate).rstrip())
    elif format in ("markdown", "md"):
        from rich.markdown import Markdown

        aggregate = []
        for component in components:
            if isinstance(component, str):
                aggregate.append(component)
            else:
                # Ignore style for now :(
                aggregate.append(component[0])

        return Markdown("".join(aggregate))
    elif format in ("restructuredtext", "rst"):
        from rich_rst import RestructuredText

        aggregate = []
        for component in components:
            if isinstance(component, str):
                aggregate.append(component)
            else:
                # Ignore style for now :(
                aggregate.append(component[0])
        return RestructuredText("".join(aggregate))
    elif format == "rich":
        from rich.text import Text

        def walk_components():
            for component in components:
                if isinstance(component, str):
                    yield Text.from_markup(component.rstrip())
                else:
                    yield Text.from_markup(component[0].rstrip(), style=component[1])

        return Text().join(walk_components())
    else:
        raise ValueError(f'Unknown help_format "{format}"')


def _get_choices(type_: Type, name_transform: Callable[[str], str]) -> str:
    get_choices = partial(_get_choices, name_transform=name_transform)
    choices: str = ""
    _origin = get_origin(type_)
    if isclass(type_) and issubclass(type_, Enum):
        choices = ",".join(name_transform(x.name) for x in type_)
    elif cyclopts.utils.is_union(_origin):
        inner_choices = [get_choices(inner) for inner in get_args(type_)]
        choices = ",".join(x for x in inner_choices if x)
    elif _origin is Literal:
        choices = ",".join(str(x) for x in get_args(type_))
    elif _origin in (list, set, tuple):
        args = get_args(type_)
        if len(args) == 1 or (_origin is tuple and len(args) == 2 and args[1] is Ellipsis):
            choices = get_choices(args[0])
    elif TypeAliasType is not None and isinstance(type_, TypeAliasType):
        choices = get_choices(type_.__value__)
    return choices


def create_parameter_help_panel(
    group: "Group",
    iparams,
    cparams: List[Parameter],
    format: str,
) -> HelpPanel:
    help_panel = HelpPanel(format="parameter", title=group.name, description=format_str(group.help, format=format))
    icparams = [(ip, cp) for ip, cp in zip(iparams, cparams) if cp.show]

    if not icparams:
        return help_panel

    iparams, cparams = (list(x) for x in zip(*icparams))

    def help_append(text, style=""):
        if help_components:
            text = " " + text
        if style:
            help_components.append((text, style))
        else:
            help_components.append(text)

    for iparam, cparam in icparams:
        assert cparam.name is not None
        assert cparam.name_transform is not None
        type_ = get_hint_parameter(iparam)[0]
        options = list(cparam.name)
        options.extend(cparam.get_negatives(type_, *options))

        # Add an all-uppercase name if it's an argument
        if iparam.kind in (iparam.POSITIONAL_ONLY, iparam.POSITIONAL_OR_KEYWORD):
            arg_name = options[0].lstrip("-").upper()
            if arg_name != options[0]:
                options = [arg_name, *options]

        short_options, long_options = [], []
        for option in options:
            if _is_short(option):
                short_options.append(option)
            else:
                long_options.append(option)

        help_components = []

        if cparam.help:
            help_append(cparam.help)

        if cparam.show_choices:
            choices = _get_choices(type_, cparam.name_transform)
            if choices:
                help_append(rf"[choices: {choices}]", "dim")

        if cparam.show_env_var and cparam.env_var:
            env_vars = " ".join(cparam.env_var)
            help_append(rf"[env var: {env_vars}]", "dim")

        if cparam.show_default or (
            cparam.show_default is None and iparam.default not in {None, inspect.Parameter.empty}
        ):
            default = ""
            if isclass(type_) and issubclass(type_, Enum):
                default = cparam.name_transform(iparam.default.name)
            else:
                default = iparam.default

            help_append(rf"[default: {default}]", "dim")

        if cparam.required:
            help_append(r"[required]", "dim red")

        # populate row
        help_panel.entries.append(
            HelpEntry(
                name=",".join(long_options),
                description=format_str(*help_components, format=format),
                short=",".join(short_options),
                required=bool(cparam.required),
            )
        )

    return help_panel


def format_command_entries(apps: Iterable["App"], format: str) -> List:
    entries = []
    for app in apps:
        short_names, long_names = [], []
        for name in app.name:
            short_names.append(name) if _is_short(name) else long_names.append(name)
        entry = HelpEntry(
            name=",".join(long_names),
            short=",".join(short_names),
            description=format_str(docstring_parse(app.help).short_description or "", format=format),
        )
        if entry not in entries:
            entries.append(entry)
    return entries


def resolve_help_format(app_chain: Optional[Iterable["App"]]) -> str:
    # Resolve help_format; None fallsback to parent; non-None overwrites parent.
    format_ = "restructuredtext"
    if app_chain is None:
        return format_
    for app in app_chain:
        if app.help_format is not None:
            format_ = app.help_format
    return format_


def resolve_version_format(app_chain: Optional[Iterable["App"]]) -> str:
    # Resolve version_format; None fallsback to parent; non-None overwrites parent.
    format_ = resolve_help_format(app_chain)
    if app_chain is None:
        return format_
    for app in app_chain:
        if app.version_format is not None:
            format_ = app.version_format
    return format_



import sys
from pathlib import Path

from cyclopts import validators
from cyclopts._convert import convert
from cyclopts.parameter import Parameter

if sys.version_info < (3, 9):
    from typing_extensions import Annotated  # pragma: no cover
else:
    from typing import Annotated  # pragma: no cover

__all__ = [
    # Path
    "ExistingPath",
    "ExistingFile",
    "ExistingDirectory",
    "Directory",
    "File",
    "ResolvedExistingPath",
    "ResolvedExistingFile",
    "ResolvedExistingDirectory",
    "ResolvedDirectory",
    "ResolvedFile",
    # Number
    "PositiveFloat",
    "NonNegativeFloat",
    "NegativeFloat",
    "NonPositiveFloat",
    "PositiveInt",
    "NonNegativeInt",
    "NegativeInt",
    "NonPositiveInt",
]


########
# Path #
########
def _path_resolve_converter(type_, *args):
    return convert(type_, *args, converter=lambda _, x: Path(x).resolve())


ExistingPath = Annotated[Path, Parameter(validator=validators.Path(exists=True))]
"A :class:`~pathlib.Path` file or directory that **must** exist."

ResolvedPath = Annotated[Path, Parameter(converter=_path_resolve_converter)]
"A :class:`~pathlib.Path` file or directory. :meth:`~pathlib.Path.resolve` is invoked prior to returning the path."
ResolvedExistingPath = Annotated[ExistingPath, Parameter(converter=_path_resolve_converter)]
"A :class:`~pathlib.Path` file or directory that **must** exist. :meth:`~pathlib.Path.resolve` is invoked prior to returning the path."

Directory = Annotated[Path, Parameter(validator=validators.Path(file_okay=False))]
"A :class:`~pathlib.Path` that **must** be a directory (or not exist)."
ExistingDirectory = Annotated[Path, Parameter(validator=validators.Path(exists=True, file_okay=False))]
"A :class:`~pathlib.Path` directory that **must** exist."
ResolvedDirectory = Annotated[Directory, Parameter(converter=_path_resolve_converter)]
"A :class:`~pathlib.Path` directory. :meth:`~pathlib.Path.resolve` is invoked prior to returning the path."
ResolvedExistingDirectory = Annotated[ExistingDirectory, Parameter(converter=_path_resolve_converter)]
"A :class:`~pathlib.Path` directory that **must** exist. :meth:`~pathlib.Path.resolve` is invoked prior to returning the path."

File = Annotated[Path, Parameter(validator=validators.Path(dir_okay=False))]
"A :class:`~pathlib.File` that **must** be a file (or not exist)."
ExistingFile = Annotated[Path, Parameter(validator=validators.Path(exists=True, dir_okay=False))]
"A :class:`~pathlib.Path` file that **must** exist."
ResolvedFile = Annotated[File, Parameter(converter=_path_resolve_converter)]
"A :class:`~pathlib.Path` file. :meth:`~pathlib.Path.resolve` is invoked prior to returning the path."
ResolvedExistingFile = Annotated[ExistingFile, Parameter(converter=_path_resolve_converter)]
"A :class:`~pathlib.Path` file that **must** exist. :meth:`~pathlib.Path.resolve` is invoked prior to returning the path."


##########
# Number #
##########
# foo
PositiveFloat = Annotated[float, Parameter(validator=validators.Number(gt=0))]
"A float that **must** be ``>0``."
NonNegativeFloat = Annotated[float, Parameter(validator=validators.Number(gte=0))]
"A float that **must** be ``>=0``."
NegativeFloat = Annotated[float, Parameter(validator=validators.Number(lt=0))]
"A float that **must** be ``<0``."
NonPositiveFloat = Annotated[float, Parameter(validator=validators.Number(lte=0))]
"A float that **must** be ``<=0``."

PositiveInt = Annotated[int, Parameter(validator=validators.Number(gt=0))]
"An int that **must** be ``>0``."
NonNegativeInt = Annotated[int, Parameter(validator=validators.Number(gte=0))]
"An int that **must** be ``>=0``."
NegativeInt = Annotated[int, Parameter(validator=validators.Number(lt=0))]
"An int that **must** be ``<0``."
NonPositiveInt = Annotated[int, Parameter(validator=validators.Number(lte=0))]
"An int that **must** be ``<=0``."



import pathlib
from typing import Type

from attrs import frozen


@frozen(kw_only=True)
class Path:
    """Assertions on properties of :class:`pathlib.Path`."""

    exists: bool = False
    """If ``True``, specified path **must** exist. Defaults to ``False``."""

    file_okay: bool = True
    """
    If ``True``, specified path may be a file.
    If ``False``, then files are not allowed.
    Defaults to ``True``.
    """

    dir_okay: bool = True
    """
    If ``True``, specified path may be a directory.
    If ``False``, then directories are not allowed.
    Defaults to ``True``.
    """

    def __attrs_post_init__(self):
        if not self.file_okay and not self.dir_okay:
            raise ValueError("file_okay and dir_okay cannot both be False.")

    def __call__(self, type_: Type, path: pathlib.Path):
        if not isinstance(path, pathlib.Path):
            raise TypeError

        if path.exists():
            if not self.file_okay and path.is_file():
                raise ValueError(f"Only directory input is allowed but {path} is a file.")

            if not self.dir_okay and path.is_dir():
                raise ValueError(f"Only file input is allowed but {path} is a directory.")
        elif self.exists:
            raise ValueError(f"{path} does not exist.")



import inspect
import os
import sys
import traceback
from contextlib import suppress
from copy import copy
from functools import partial
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Literal,
    Mapping,
    Optional,
    Tuple,
    TypeVar,
    Union,
    overload,
)

from attrs import define, field

import cyclopts.utils
from cyclopts.bind import create_bound_arguments, normalize_tokens
from cyclopts.exceptions import (
    CommandCollisionError,
    CycloptsError,
    InvalidCommandError,
    UnusedCliTokensError,
    ValidationError,
    format_cyclopts_error,
)
from cyclopts.group import Group, GroupConverter, sort_groups
from cyclopts.group_extractors import groups_from_app, inverse_groups_from_app
from cyclopts.help import (
    HelpPanel,
    create_parameter_help_panel,
    format_command_entries,
    format_doc,
    format_str,
    format_usage,
    resolve_help_format,
    resolve_version_format,
)
from cyclopts.parameter import Parameter, validate_command
from cyclopts.protocols import Dispatcher
from cyclopts.resolve import ResolvedCommand
from cyclopts.utils import (
    default_name_transform,
    optional_to_tuple_converter,
    to_list_converter,
    to_tuple_converter,
)

T = TypeVar("T", bound=Callable)


if sys.version_info < (3, 9):
    from typing_extensions import Annotated
else:
    from typing import Annotated


with suppress(ImportError):
    # By importing, makes things like the arrow-keys work.
    import readline  # Not available on windows

if TYPE_CHECKING:
    from rich.console import Console


class _CannotDeriveCallingModuleNameError(Exception):
    pass


def _get_root_module_name():
    """Get the calling package name from the call-stack."""
    for elem in inspect.stack():
        module = inspect.getmodule(elem.frame)
        if module is None:
            continue
        root_module_name = module.__name__.split(".")[0]
        if root_module_name == "cyclopts":
            continue
        return root_module_name

    raise _CannotDeriveCallingModuleNameError  # pragma: no cover


def _default_version(default="0.0.0") -> str:
    """Attempts to get the calling code's version.

    Returns
    -------
    version: str
        ``default`` if it cannot determine version.
    """
    import importlib

    if sys.version_info < (3, 10):  # pragma: no cover
        from importlib_metadata import PackageNotFoundError
        from importlib_metadata import version as importlib_metadata_version
    else:  # pragma: no cover
        from importlib.metadata import PackageNotFoundError
        from importlib.metadata import version as importlib_metadata_version

    try:
        root_module_name = _get_root_module_name()
    except _CannotDeriveCallingModuleNameError:  # pragma: no cover
        return default

    # Attempt to get the Distribution Package’s version number.
    try:
        return importlib_metadata_version(root_module_name)
    except PackageNotFoundError:
        pass

    # Attempt packagename.__version__
    # Not sure if this is redundant with ``importlib.metadata``,
    # but there's no real harm in checking.
    try:
        module = importlib.import_module(root_module_name)
        return module.__version__
    except (ImportError, AttributeError):
        pass

    # Final fallback
    return default


def _validate_default_command(x):
    if isinstance(x, App):
        raise TypeError("Cannot register a sub-App to default.")
    return x


def _combined_meta_command_mapping(app):
    """Return a copied and combined mapping containing app and meta-app commands."""
    command_mapping = copy(app._commands)
    while (app := app._meta) and app._commands:
        command_mapping.update(app._commands)
    return command_mapping


def _get_command_groups(parent_app, child_app):
    """Extract out the command groups from the ``parent_app`` for a given ``child_app``."""
    return next(x for x in inverse_groups_from_app(parent_app) if x[0] is child_app)[1]


def resolve_default_parameter_from_apps(apps) -> Parameter:
    """The default_parameter resolution depends on the parent-child path traversed."""
    cparams = []
    for parent_app, child_app in zip(apps[:-1], apps[1:]):
        # child_app could be a command of parent_app.meta
        if parent_app._meta and child_app in parent_app._meta._commands.values():
            cparams = []  # meta-apps do NOT inherit from their parenting app.
            parent_app = parent_app._meta

        groups = _get_command_groups(parent_app, child_app)
        cparams.extend([group.default_parameter for group in groups])
        cparams.append(parent_app.default_parameter)

    cparams.append(apps[-1].default_parameter)

    return Parameter.combine(*cparams)


def walk_metas(app):
    # Iterates from deepest to shallowest meta-apps
    meta_list = [app]  # shallowest to deepest
    meta = app
    while (meta := meta._meta) and meta.default_command:
        meta_list.append(meta)
    yield from reversed(meta_list)


@define
class App:
    # This can ONLY ever be Tuple[str, ...] due to converter.
    # The other types is to make mypy happy for Cyclopts users.
    _name: Union[None, str, Tuple[str, ...]] = field(default=None, alias="name", converter=optional_to_tuple_converter)

    _help: Optional[str] = field(default=None, alias="help")

    usage: Optional[str] = field(default=None)

    # Everything below must be kw_only

    default_command: Optional[Callable] = field(default=None, converter=_validate_default_command, kw_only=True)
    default_parameter: Optional[Parameter] = field(default=None, kw_only=True)

    # This can ONLY ever be None or Tuple[Callable, ...]
    _config: Union[None, Callable, Iterable[Callable]] = field(
        default=None,
        alias="config",
        converter=optional_to_tuple_converter,
        kw_only=True,
    )

    version: Union[None, str, Callable[..., str]] = field(factory=_default_version, kw_only=True)
    # This can ONLY ever be a Tuple[str, ...]
    _version_flags: Union[str, Iterable[str]] = field(
        default=["--version"],
        converter=to_tuple_converter,
        alias="version_flags",
        kw_only=True,
    )

    show: bool = field(default=True, kw_only=True)

    console: Optional["Console"] = field(default=None, kw_only=True)

    # This can ONLY ever be a Tuple[str, ...]
    _help_flags: Union[str, Iterable[str]] = field(
        default=["--help", "-h"],
        converter=to_tuple_converter,
        alias="help_flags",
        kw_only=True,
    )
    help_format: Optional[
        Literal[
            "markdown",
            "md",
            "plaintext",
            "restructuredtext",
            "rst",
            "rich",
        ]
    ] = field(default=None, kw_only=True)

    version_format: Optional[
        Literal[
            "markdown",
            "md",
            "plaintext",
            "restructuredtext",
            "rst",
            "rich",
        ]
    ] = field(default=None, kw_only=True)

    # This can ONLY ever be Tuple[Union[Group, str], ...] due to converter.
    # The other types is to make mypy happy for Cyclopts users.
    group: Union[Group, str, Tuple[Union[Group, str], ...]] = field(
        default=None, converter=to_tuple_converter, kw_only=True
    )

    # This can ONLY ever be a Group
    group_arguments: Union[Group, str, None] = field(
        default=None,
        converter=GroupConverter(Group.create_default_arguments()),
        kw_only=True,
    )
    # This can ONLY ever be a Group
    group_parameters: Union[Group, str, None] = field(
        default=None,
        converter=GroupConverter(Group.create_default_parameters()),
        kw_only=True,
    )
    # This can ONLY ever be a Group
    group_commands: Union[Group, str, None] = field(
        default=None,
        converter=GroupConverter(Group.create_default_commands()),
        kw_only=True,
    )

    converter: Optional[Callable[..., Mapping[str, Any]]] = field(default=None, kw_only=True)
    validator: List[Callable[..., Any]] = field(default=None, converter=to_list_converter, kw_only=True)

    _name_transform: Optional[Callable[[str], str]] = field(
        default=None,
        alias="name_transform",
        kw_only=True,
    )

    ######################
    # Private Attributes #
    ######################
    # Maps CLI-name of a command to a function handle.
    _commands: Dict[str, "App"] = field(init=False, factory=dict)

    _parents: List["App"] = field(init=False, factory=list)

    _meta: "App" = field(init=False, default=None)
    _meta_parent: "App" = field(init=False, default=None)

    def __attrs_post_init__(self):
        # Trigger the setters
        self.help_flags = self._help_flags
        self.version_flags = self._version_flags

    ###########
    # Methods #
    ###########
    def _delete_commands(self, commands: Iterable[str], default=None):
        """Safely delete commands.

        Will **not** raise an exception if command(s) do not exist.

        Parameters
        ----------
        commands: Iterable[str, ...]
            Strings of commands to delete.
        """
        # Remove all the old version-flag commands.
        for command in commands:
            with suppress(KeyError):
                if default:
                    if self[command].default == self.version_print:
                        del self[command]
                else:
                    del self[command]

    @property
    def version_flags(self):
        return self._version_flags

    @version_flags.setter
    def version_flags(self, value):
        self._version_flags = value
        self._delete_commands(self._version_flags, default=self.version_print)
        if self._version_flags:
            self.command(
                self.version_print,
                name=self._version_flags,
                help_flags=[],
                version_flags=[],
                help="Display application version.",
            )

    @property
    def help_flags(self):
        return self._help_flags

    @help_flags.setter
    def help_flags(self, value):
        self._help_flags = value
        self._delete_commands(self._help_flags, default=self.help_print)
        if self._help_flags:
            self.command(
                self.help_print,
                name=self._help_flags,
                help_flags=[],
                version_flags=[],
                help="Display this message and exit.",
            )

    @property
    def name(self) -> Tuple[str, ...]:
        """Application name(s). Dynamically derived if not previously set."""
        if self._name:
            return self._name  # pyright: ignore[reportReturnType]
        elif self.default_command is None:
            name = Path(sys.argv[0]).name
            if name == "__main__.py":
                name = _get_root_module_name()
            return (name,)
        else:
            return (self.name_transform(self.default_command.__name__),)

    @property
    def config(self) -> Tuple[str, ...]:
        if self._config is None and self._meta_parent is not None:
            return self._meta_parent.config
        else:
            return self._config or ()  # pyright: ignore[reportReturnType]

    @config.setter
    def config(self, value):
        self._config = value

    @property
    def help(self) -> str:
        if self._help is not None:
            return self._help
        elif self.default_command is None:
            # Try and fallback to a meta-app docstring.
            if self._meta is None:
                return ""
            else:
                return self.meta.help
        elif self.default_command.__doc__ is None:
            return ""
        else:
            return self.default_command.__doc__

    @help.setter
    def help(self, value):
        self._help = value

    @property
    def name_transform(self):
        return self._name_transform if self._name_transform else default_name_transform

    @name_transform.setter
    def name_transform(self, value):
        self._name_transform = value

    def version_print(
        self,
        console: Optional["Console"] = None,
    ) -> None:
        """Print the application version.

        Parameters
        ----------
        console: rich.console.Console
            Console to print version string to.
            If not provided, follows the resolution order defined in :attr:`App.console`.

        """
        console = self._resolve_console(None, console)
        version_format = resolve_version_format([self])

        version_raw = self.version() if callable(self.version) else self.version

        if version_raw is None:
            version_raw = "0.0.0"

        version_formatted = format_str(version_raw, format=version_format)
        console.print(version_formatted)

    def __getitem__(self, key: str) -> "App":
        """Get the subapp from a command string.

        All commands get registered to Cyclopts as subapps.
        The actual function handler is at ``app[key].default_command``.
        """
        if self._meta:
            with suppress(KeyError):
                return self.meta[key]
        return self._commands[key]

    def __delitem__(self, key: str):
        del self._commands[key]

    def __contains__(self, k: str) -> bool:
        if k in self._commands:
            return True
        if self._meta_parent:
            return k in self._meta_parent
        return False

    def __iter__(self) -> Iterator[str]:
        """Iterate over command & meta command names."""
        for k in self._commands:
            yield k
        if self._meta_parent:
            for k in self._meta_parent:
                yield k

    @property
    def meta(self) -> "App":
        if self._meta is None:
            self._meta = type(self)(
                help_flags=copy(self.help_flags),
                version_flags=copy(self.version_flags),
                group_commands=copy(self.group_commands),
                group_arguments=copy(self.group_arguments),
                group_parameters=copy(self.group_parameters),
            )
            self._meta._meta_parent = self
        return self._meta

    def parse_commands(
        self,
        tokens: Union[None, str, Iterable[str]] = None,
    ) -> Tuple[Tuple[str, ...], Tuple["App", ...], List[str]]:
        """Extract out the command tokens from a command.

        Parameters
        ----------
        tokens: Union[None, str, Iterable[str]]
            Either a string, or a list of strings to launch a command.
            Defaults to ``sys.argv[1:]``

        Returns
        -------
        List[str]
            Tokens that are interpreted as a valid command chain.
        List[App]
            The associated :class:`App` object with each of those tokens.
        List[str]
            The remaining non-command tokens.
        """
        tokens = normalize_tokens(tokens)

        command_chain = []
        app = self
        apps = [app]
        unused_tokens = tokens

        command_mapping = _combined_meta_command_mapping(app)

        for i, token in enumerate(tokens):
            try:
                app = command_mapping[token]
                apps.append(app)
                unused_tokens = tokens[i + 1 :]
            except KeyError:
                break
            command_chain.append(token)
            command_mapping = _combined_meta_command_mapping(app)

        return tuple(command_chain), tuple(apps), unused_tokens

    # This overload is used in code like:
    #
    # @app.command
    # def my_command(foo: str):
    #   ...
    @overload
    def command(
        self,
        obj: T,
        name: Union[None, str, Iterable[str]] = None,
        **kwargs: object,
    ) -> T: ...

    # This overload is used in code like:
    #
    # @app.command(name="bar")
    # def my_command(foo: str):
    #   ...
    @overload
    def command(
        self,
        obj: None = None,
        name: Union[None, str, Iterable[str]] = None,
        **kwargs: object,
    ) -> Callable[[T], T]: ...

    def command(
        self,
        obj: Optional[T] = None,
        name: Union[None, str, Iterable[str]] = None,
        **kwargs: object,
    ) -> Union[T, Callable[[T], T]]:
        """Decorator to register a function as a CLI command.

        Parameters
        ----------
        obj: Optional[Callable]
            Function or :class:`App` to be registered as a command.
        name: Union[None, str, Iterable[str]]
            Name(s) to register the ``obj`` to.
            If not provided, defaults to:

            * If registering an :class:`App`, then the app's name.
            * If registering a function, then the function's name.
        `**kwargs`
            Any argument that :class:`App` can take.
        """
        if obj is None:  # Called ``@app.command(...)``
            return partial(self.command, name=name, **kwargs)  # pyright: ignore[reportReturnType]

        if isinstance(obj, App):
            app = obj

            if app._name is None and name is None:
                raise ValueError("Sub-app MUST have a name specified.")

            if kwargs:
                raise ValueError("Cannot supplied additional configuration when registering a sub-App.")
        else:
            validate_command(obj)
            kwargs.setdefault("help_flags", [])
            kwargs.setdefault("version_flags", [])
            if "group_commands" not in kwargs:
                kwargs["group_commands"] = copy(self.group_commands)
            if "group_parameters" not in kwargs:
                kwargs["group_parameters"] = copy(self.group_parameters)
            if "group_arguments" not in kwargs:
                kwargs["group_arguments"] = copy(self.group_arguments)
            app = App(default_command=obj, **kwargs)  # pyright: ignore
            # app.name is handled below

        if app._name_transform is None:
            app.name_transform = self.name_transform

        if name is None:
            name = app.name
        else:
            app._name = name  # pyright: ignore[reportAttributeAccessIssue]

        for n in to_tuple_converter(name):
            if n in self:
                raise CommandCollisionError(f'Command "{n}" already registered.')

            # Warning: app._name may not align with command name
            self._commands[n] = app

        app._parents.append(self)

        return obj  # pyright: ignore[reportReturnType]

    # This overload is used in code like:
    #
    # @app.default
    # def my_command(foo: str):
    #   ...
    @overload
    def default(
        self,
        obj: T,
        *,
        converter: Optional[Callable[..., Mapping[str, Any]]] = None,
        validator: Optional[Callable[..., Any]] = None,
    ) -> T: ...

    # This overload is used in code like:
    #
    # @app.default()
    # def my_command(foo: str):
    #   ...
    @overload
    def default(
        self,
        obj: None = None,
        *,
        converter: Optional[Callable[..., Mapping[str, Any]]] = None,
        validator: Optional[Callable[..., Any]] = None,
    ) -> Callable[[T], T]: ...

    def default(
        self,
        obj: Optional[T] = None,
        *,
        converter: Optional[Callable[..., Mapping[str, Any]]] = None,
        validator: Optional[Callable[..., Any]] = None,
    ) -> Union[T, Callable[[T], T]]:
        """Decorator to register a function as the default action handler."""
        if obj is None:  # Called ``@app.default_command(...)``
            return partial(self.default, converter=converter, validator=validator)  # pyright: ignore[reportReturnType]

        if isinstance(obj, App):  # Registering a sub-App
            raise TypeError("Cannot register a sub-App to default.")

        if self.default_command is not None:
            raise CommandCollisionError(f"Default command previously set to {self.default_command}.")

        validate_command(obj)
        self.default_command = obj
        if converter:
            self.converter = converter
        if validator:
            self.validator = validator  # pyright: ignore[reportAttributeAccessIssue]
        return obj

    def parse_known_args(
        self,
        tokens: Union[None, str, Iterable[str]] = None,
        *,
        console: Optional["Console"] = None,
    ) -> Tuple[Callable, inspect.BoundArguments, List[str]]:
        """Interpret arguments into a function, :class:`~inspect.BoundArguments`, and any remaining unknown tokens.

        Parameters
        ----------
        tokens: Union[None, str, Iterable[str]]
            Either a string, or a list of strings to launch a command.
            Defaults to ``sys.argv[1:]``
        console: rich.console.Console
            Console to print help and runtime Cyclopts errors.
            If not provided, follows the resolution order defined in :attr:`App.console`.

        Returns
        -------
        command: Callable
            Bare function to execute.

        bound: inspect.BoundArguments
            Bound arguments for ``command``.

        unused_tokens: List[str]
            Any remaining CLI tokens that didn't get parsed for ``command``.
        """
        tokens = normalize_tokens(tokens)

        meta_parent = self

        command_chain, apps, unused_tokens = self.parse_commands(tokens)
        command_app = apps[-1]

        try:
            parent_app = apps[-2]
        except IndexError:
            parent_app = None

        config: Tuple[Callable, ...] = ()
        for app in reversed(apps):
            if app.config:
                config = app.config  # pyright: ignore[reportAssignmentType]
                break
        config = tuple(partial(x, apps, command_chain) for x in config)

        # Special flags (help/version) get intercepted by the root app.
        # Special flags are allows to be **anywhere** in the token stream.

        for help_flag in self.help_flags:
            try:
                help_flag_index = tokens.index(help_flag)
                break
            except ValueError:
                pass
        else:
            help_flag_index = None

        if help_flag_index is not None:
            tokens.pop(help_flag_index)
            command = self.help_print
            while meta_parent := meta_parent._meta_parent:
                command = meta_parent.help_print
            bound = cyclopts.utils.signature(command).bind(tokens, console=console)
            unused_tokens = []
        elif any(flag in tokens for flag in self.version_flags):
            # Version
            command = self.version_print
            while meta_parent := meta_parent._meta_parent:
                command = meta_parent.version_print
            bound = cyclopts.utils.signature(command).bind()
            unused_tokens = []
        else:
            try:
                if command_app.default_command:
                    command = command_app.default_command
                    resolved_command = self._resolve_command(tokens, parse_docstring=False)
                    # We want the resolved group that ``app`` belongs to.
                    command_groups = [] if parent_app is None else _get_command_groups(parent_app, command_app)

                    bound, unused_tokens = create_bound_arguments(resolved_command, unused_tokens, config)
                    try:
                        if command_app.converter:
                            bound.arguments = command_app.converter(**bound.arguments)  # pyright: ignore[reportAttributeAccessIssue]
                        for command_group in command_groups:
                            if command_group.converter:
                                bound.arguments = command_group.converter(**bound.arguments)
                        for validator in command_app.validator:
                            validator(**bound.arguments)
                        for command_group in command_groups:
                            for validator in command_group.validator:  # pyright: ignore
                                validator(**bound.arguments)
                    except (AssertionError, ValueError, TypeError) as e:
                        raise ValidationError(
                            value=e.args[0] if e.args else "",
                            group=command_group,  # pyright: ignore[reportPossiblyUnboundVariable]
                        ) from e

                else:
                    if unused_tokens:
                        raise InvalidCommandError(unused_tokens=unused_tokens)
                    else:
                        # Running the application with no arguments and no registered
                        # ``default_command`` will default to ``help_print``.
                        command = self.help_print
                        bound = cyclopts.utils.signature(command).bind(tokens=tokens, console=console)
                        unused_tokens = []
            except CycloptsError as e:
                e.app = command_app
                if command_chain:
                    e.command_chain = command_chain
                if e.console is None:
                    e.console = self._resolve_console(tokens, console)
                raise

        return command, bound, unused_tokens

    def parse_args(
        self,
        tokens: Union[None, str, Iterable[str]] = None,
        *,
        console: Optional["Console"] = None,
        print_error: bool = True,
        exit_on_error: bool = True,
        verbose: bool = False,
    ) -> Tuple[Callable, inspect.BoundArguments]:
        """Interpret arguments into a function and :class:`~inspect.BoundArguments`.

        Raises
        ------
        UnusedCliTokensError
            If any tokens remain after parsing.

        Parameters
        ----------
        tokens: Union[None, str, Iterable[str]]
            Either a string, or a list of strings to launch a command.
            Defaults to ``sys.argv[1:]``.
        console: rich.console.Console
            Console to print help and runtime Cyclopts errors.
            If not provided, follows the resolution order defined in :attr:`App.console`.
        print_error: bool
            Print a rich-formatted error on error.
            Defaults to ``True``.
        exit_on_error: bool
            If there is an error parsing the CLI tokens invoke ``sys.exit(1)``.
            Otherwise, continue to raise the exception.
            Defaults to ``True``.
        verbose: bool
            Populate exception strings with more information intended for developers.
            Defaults to ``False``.

        Returns
        -------
        command: Callable
            Function associated with command action.

        bound: inspect.BoundArguments
            Parsed and converted ``args`` and ``kwargs`` to be used when calling ``command``.
        """
        tokens = normalize_tokens(tokens)

        # Normal parsing
        try:
            command, bound, unused_tokens = self.parse_known_args(tokens, console=console)
            if unused_tokens:
                raise UnusedCliTokensError(
                    target=command,
                    unused_tokens=unused_tokens,
                )
        except CycloptsError as e:
            e.verbose = verbose
            e.root_input_tokens = tokens

            if e.console is None:
                e.console = self._resolve_console(tokens, console)
            if print_error:
                assert e.console
                e.console.print(format_cyclopts_error(e))
            if exit_on_error:
                sys.exit(1)
            raise

        return command, bound

    def __call__(
        self,
        tokens: Union[None, str, Iterable[str]] = None,
        *,
        console: Optional["Console"] = None,
        print_error: bool = True,
        exit_on_error: bool = True,
        verbose: bool = False,
    ):
        """Interprets and executes a command.

        Parameters
        ----------
        tokens : Union[None, str, Iterable[str]]
            Either a string, or a list of strings to launch a command.
            Defaults to ``sys.argv[1:]``.
        console: rich.console.Console
            Console to print help and runtime Cyclopts errors.
            If not provided, follows the resolution order defined in :attr:`App.console`.
        print_error: bool
            Print a rich-formatted error on error.
            Defaults to ``True``.
        exit_on_error: bool
            If there is an error parsing the CLI tokens invoke ``sys.exit(1)``.
            Otherwise, continue to raise the exception.
            Defaults to ``True``.
        verbose: bool
            Populate exception strings with more information intended for developers.
            Defaults to ``False``.

        Returns
        -------
        return_value: Any
            The value the parsed command handler returns.
        """
        tokens = normalize_tokens(tokens)
        command, bound = self.parse_args(
            tokens,
            console=console,
            print_error=print_error,
            exit_on_error=exit_on_error,
            verbose=verbose,
        )
        try:
            if inspect.iscoroutinefunction(command):
                import asyncio

                return asyncio.run(command(*bound.args, **bound.kwargs))
            else:
                return command(*bound.args, **bound.kwargs)
        except Exception as e:
            try:
                from pydantic import ValidationError as PydanticValidationError
            except ImportError:
                PydanticValidationError = None  # noqa: N806

            if PydanticValidationError is not None and isinstance(e, PydanticValidationError):
                if print_error:
                    console = self._resolve_console(tokens, console)
                    console.print(format_cyclopts_error(e))

                if exit_on_error:
                    sys.exit(1)
            raise

    def _resolve_console(
        self, tokens: Union[None, str, Iterable[str]], console: Optional["Console"] = None
    ) -> "Console":
        if console is not None:
            return console
        _, apps, _ = self.parse_commands(tokens)
        for app in reversed(apps):
            if app.console:
                return app.console
        from rich.console import Console

        return Console()

    def help_print(
        self,
        tokens: Annotated[Union[None, str, Iterable[str]], Parameter(show=False)] = None,
        *,
        console: Annotated[Optional["Console"], Parameter(parse=False)] = None,
    ) -> None:
        """Print the help page.

        Parameters
        ----------
        tokens: Union[None, str, Iterable[str]]
            Tokens to interpret for traversing the application command structure.
            If not provided, defaults to ``sys.argv``.
        console: rich.console.Console
            Console to print help and runtime Cyclopts errors.
            If not provided, follows the resolution order defined in :attr:`App.console`.
        """
        tokens = normalize_tokens(tokens)

        command_chain, apps, _ = self.parse_commands(tokens)
        executing_app = apps[-1]

        console = self._resolve_console(tokens, console)

        # Print the:
        #    my-app command COMMAND [ARGS] [OPTIONS]
        if executing_app.usage is None:
            console.print(format_usage(self, command_chain))
        elif executing_app.usage:  # i.e. skip empty-string.
            console.print(executing_app.usage + "\n")

        # Print the App/Command's Doc String.
        help_format = resolve_help_format(apps)
        console.print(format_doc(self, executing_app, help_format))

        for help_panel in self._assemble_help_panels(tokens, help_format):
            console.print(help_panel)

    def _resolve_command(
        self,
        tokens: Union[None, str, Iterable[str]] = None,
        parse_docstring: bool = True,
    ) -> ResolvedCommand:
        _, apps, _ = self.parse_commands(tokens)

        if not apps[-1].default_command:
            raise InvalidCommandError

        assert isinstance(apps[-1].group_arguments, Group)
        assert isinstance(apps[-1].group_parameters, Group)

        resolved_command = ResolvedCommand(
            apps[-1].default_command,
            resolve_default_parameter_from_apps(apps),
            apps[-1].group_arguments,
            apps[-1].group_parameters,
            parse_docstring=parse_docstring,
        )
        return resolved_command

    def _assemble_help_panels(
        self,
        tokens: Union[None, str, Iterable[str]],
        help_format,
    ) -> List[HelpPanel]:
        from rich.console import Group as RichGroup
        from rich.console import NewLine

        _, apps, _ = self.parse_commands(tokens)

        help_format = resolve_help_format(apps)

        panels: Dict[str, Tuple[Group, HelpPanel]] = {}
        # Handle commands first; there's an off chance they may be "upgraded"
        # to an argument/parameter panel.
        for subapp in walk_metas(apps[-1]):
            # Handle Commands
            for group, elements in groups_from_app(subapp):
                if not group.show:
                    continue

                try:
                    _, command_panel = panels[group.name]
                except KeyError:
                    command_panel = HelpPanel(
                        format="command",
                        title=group.name,
                    )
                    panels[group.name] = (group, command_panel)

                if group.help:
                    group_help = format_str(group.help, format=help_format)

                    if command_panel.description:
                        command_panel.description = RichGroup(command_panel.description, NewLine(), group_help)
                    else:
                        command_panel.description = group_help

                command_panel.entries.extend(format_command_entries(elements, format=help_format))

        # Handle Arguments/Parameters
        for subapp in walk_metas(apps[-1]):
            if not subapp.default_command:
                continue
            command = ResolvedCommand(
                subapp.default_command,
                resolve_default_parameter_from_apps(apps),
                subapp.group_arguments,
                subapp.group_parameters,
            )
            for group, iparams in command.groups_iparams:
                if not group.show:
                    continue
                cparams = [command.iparam_to_cparam[x] for x in iparams]
                try:
                    _, existing_panel = panels[group.name]
                except KeyError:
                    existing_panel = None
                new_panel = create_parameter_help_panel(group, iparams, cparams, help_format)

                if existing_panel:
                    # An imperfect merging process
                    existing_panel.format = "parameter"
                    existing_panel.entries = new_panel.entries + existing_panel.entries  # Commands go last
                    if new_panel.description:
                        if existing_panel.description:
                            existing_panel.description = RichGroup(
                                existing_panel.description, NewLine(), new_panel.description
                            )
                        else:
                            existing_panel.description = new_panel.description
                else:
                    panels[group.name] = (group, new_panel)

        groups = [x[0] for x in panels.values()]
        help_panels = [x[1] for x in panels.values()]

        out = []
        for help_panel in sort_groups(groups, help_panels)[1]:
            help_panel.remove_duplicates()
            if help_panel.format == "command":
                # don't sort format == "parameter" because order may matter there!
                help_panel.sort()
            out.append(help_panel)
        return out

    def interactive_shell(
        self,
        prompt: str = "$ ",
        quit: Union[None, str, Iterable[str]] = None,
        dispatcher: Optional[Dispatcher] = None,
        **kwargs,
    ) -> None:
        """Create a blocking, interactive shell.

        All registered commands can be executed in the shell.

        Parameters
        ----------
        prompt: str
            Shell prompt. Defaults to ``"$ "``.
        quit: Union[str, Iterable[str]]
            String or list of strings that will cause the shell to exit and this method to return.
            Defaults to ``["q", "quit"]``.
        dispatcher: Optional[Dispatcher]
            Optional function that subsequently invokes the command.
            The ``dispatcher`` function must have signature:

            .. code-block:: python

                def dispatcher(command: Callable, bound: inspect.BoundArguments) -> Any:
                    return command(*bound.args, **bound.kwargs)

            The above is the default dispatcher implementation.
        `**kwargs`
            Get passed along to :meth:`parse_args`.
        """
        if os.name == "posix":
            print("Interactive shell. Press Ctrl-D to exit.")
        else:  # Windows
            print("Interactive shell. Press Ctrl-Z followed by Enter to exit.")

        if quit is None:
            quit = ["q", "quit"]
        if isinstance(quit, str):
            quit = [quit]

        def default_dispatcher(command, bound):
            return command(*bound.args, **bound.kwargs)

        if dispatcher is None:
            dispatcher = default_dispatcher

        kwargs.setdefault("exit_on_error", False)

        while True:
            try:
                user_input = input(prompt)
            except EOFError:
                break

            tokens = normalize_tokens(user_input)
            if not tokens:
                continue
            if tokens[0] in quit:
                break

            try:
                command, bound = self.parse_args(tokens, **kwargs)
                dispatcher(command, bound)
            except CycloptsError:
                # Upstream ``parse_args`` already printed the error
                pass
            except Exception:
                print(traceback.format_exc())

    def __repr__(self):
        """Only shows non-default values."""
        non_defaults = {}
        for a in self.__attrs_attrs__:  # pyright: ignore[reportAttributeAccessIssue]
            if not a.init:
                continue
            v = getattr(self, a.name)
            # Compare types first because of some weird attribute issues.
            if type(v) != type(a.default) or v != a.default:  # noqa: E721
                non_defaults[a.alias] = v

        signature = ", ".join(f"{k}={v!r}" for k, v in non_defaults.items())
        return f"{type(self).__name__}({signature})"
